/*
 * kexec: Linux boots Linux
 *
 * Copyright (C) 2003,2004  Eric Biederman (ebiederm@xmission.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

.data
	.equ	MSR_K6_EFER,   0xC0000080
	.equ	EFER_LME,      0x00000100
	.equ	X86_CR4_PAE,   0x00000020
	.equ	CR0_PG,        0x80000000

	.globl setup32_start, setup32_end, setup32_size
	.globl setup32_regs
setup32_start:
_reloc = .
	.code64
	/* Compute where I am running at */
	leaq	setup32_start(%rip), %rbx

	/* Relocate the code */
	addq	%rbx, lreloc1(%rip)
	addl	%ebx, reloc1(%rip)
	addl	%ebx, reloc2(%rip)
	addl	%ebx, reloc3(%rip)

	/* Setup a gdt that should be preserved */
	/* This also acts as a serializing instruction ensuring
	 * my self modifying code works.
	 */
	lgdt	gdt(%rip)
		
	/* load the data segments */
	movw	$0x18, %ax	/* data segment */
	movw	%ax, %ds
	movw	%ax, %es
	movw	%ax, %ss
	movw	%ax, %fs
	movw	%ax, %gs

	/* Switch to 32bit compatiblity mode */
	ljmp	*lm_exit_addr(%rip)
lm_exit:
	.code32

	/* Disable paging */
	movl	%cr0, %eax
	andl	$~CR0_PG, %eax
	movl	%eax, %cr0

	/* Disable long mode */
	movl	$MSR_K6_EFER, %ecx
	rdmsr
	andl	$~EFER_LME, %eax
	wrmsr

	/* Disable PAE */
	xorl	%eax, %eax
	movl	%eax, %cr4

	/* load the data segments */
	movl	$0x18, %eax	/* data segment */
	movl	%eax, %ds
	movl	%eax, %es
	movl	%eax, %ss
	movl	%eax, %fs
	movl	%eax, %gs

	/* Load the registers */
	movl	eax - _reloc(%ebx), %eax
	movl	ecx - _reloc(%ebx), %ecx
	movl	edx - _reloc(%ebx), %edx
	movl	esi - _reloc(%ebx), %esi
	movl	edi - _reloc(%ebx), %edi
	movl	esp - _reloc(%ebx), %esp
	movl	ebp - _reloc(%ebx), %ebp

	.byte 0x8b, 0x1d	# movl ebx,  %ebx
reloc1:	.long ebx - _reloc

	nop
	nop

	.byte 0xff, 0x25	# jmpl *(eip)
reloc2:	.long eip - _reloc

	.balign 4
setup32_regs:
eax:	.long 0x00000000
ebx:	.long 0x00000000
ecx:	.long 0x00000000
edx:	.long 0x00000000
esi:	.long 0x00000000
edi:	.long 0x00000000
esp:	.long 0x00000000
ebp:	.long 0x00000000
eip:	.long 0x00000000


	.balign 16
gdt:	/* 0x00 unusable segment 
	 * 0x08 unused
	 * so use them as the gdt ptr
	 */
	.word	gdt_end - gdt - 1
lreloc1:		
	.quad	gdt - _reloc
	.word	0
			
	/* Documented linux kernel segments */
	/* 0x10 4GB flat code segment */
	.word	0xFFFF, 0x0000, 0x9A00, 0x00CF
	/* 0x18 4GB flat data segment */
	.word	0xFFFF, 0x0000, 0x9200, 0x00CF

	/* 0x20 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x28 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x30 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x38 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x40 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x48 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x50 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	/* 0x58 dummy */
	.word	0x0000, 0x0000, 0x0000, 0x000
	
	/* Segments used by the 2.5.x kernel */
	/* 0x60 4GB flat code segment */
	.word	0xFFFF, 0x0000, 0x9A00, 0x00CF
	/* 0x68 4GB flat data segment */
	.word	0xFFFF, 0x0000, 0x9200, 0x00CF
gdt_end:

lm_exit_addr:
reloc3:	       
	.long lm_exit - _reloc
	.long 0x10
	
setup32_end:
setup32_size:
	.long setup32_end - setup32_start
